import 'dart:async';

import 'package:flutter_mqtt/src/packets/packets.dart';
import 'package:flutter_xtransport/flutter_xtransport.dart';
import './packets/utils.dart';
import './packets/enum.dart';
import './packets/mqtt_byte_buffer.dart';

typedef EvtMqttPublishArrived = void Function(PublishPacket msg);

/// Mqtt client instance
class MqttClient {
  XtransportCredentials credentials;
  static MqttClient? instance;
  late XTransportTcpClient _conn;
  bool _started = false;
  bool _paused = false;
  bool _stoped = false;
  bool log = false;
  Timer? _pinger;

  /// allow Reconnect when
  bool allowReconnect;

  /// delay before reconnect
  Duration reconnectWait;

  /// custom reconnect delay
  Duration Function()? customReconnectDelayCB;
  final ConnectPacket _connectPacket = ConnectPacket();
  final _idTopic = <int, List<String>>{};
  final _subList = <String, SubscribePacket>{};
  final _subComplate = <String, void Function()>{};
  final _finishedSubCache = <String>{};
  final _dataArriveCallBack = <String, EvtMqttPublishArrived>{};
  //Events
  void Function(ConackPacket msg)? _onMqttConack;
  Future<void> Function()? _onBeforeReconnect;

  /// set ConnectPacket ping keepaliveSecond
  MqttClient withKeepalive(int seconds) {
    _connectPacket.withKeepalive(seconds);
    return this;
  }

  /// set ConnectPacket UserName and pwd
  MqttClient withAuth(String? userName, String? pwd) {
    _connectPacket.withAuth(userName, pwd);
    return this;
  }

  /// set ConnectPacket clientID
  MqttClient withClientID(String clientID) {
    _connectPacket.withClientID(clientID);
    return this;
  }

  /// set ConnectPacket clearSession default:true
  MqttClient withClearSession(bool clear) {
    _connectPacket.withClearSession(clear);
    return this;
  }

  void onBeforeReconnect(Future<void> Function()? fn) =>
      _onBeforeReconnect = fn;
  void onMqttConack(void Function(ConackPacket msg)? fn) => _onMqttConack = fn;
  Future<void> subscribe(
    String topic, {
    bool force = true,
    EvtMqttPublishArrived? onMessage,
    bool futureWaitData = false,
    // mqtt head
    bool retain = false,
    MqttQos qos = MqttQos.atMostOnce,
    bool dup = false,
    Duration? timeout,
  }) {
    var com = Completer<void>();
    Timer? _timeout;
    if (timeout != null) {
      _timeout = Timer.periodic(timeout, (timer) {
        timer.cancel();
        if (!com.isCompleted) {
          com.completeError("subcribe timeout");
        }
      });
    }
    if (onMessage != null) {
      _dataArriveCallBack[topic] = onMessage;
    }
    // handle the events
    _dataArriveCallBack[topic] = (msg) {
      //futureWaitData
      if (futureWaitData && !com.isCompleted) {
        _timeout?.cancel();
        com.complete();
      }
      onMessage?.call(msg);
    };

    if (!force && _finishedSubCache.contains(topic)) {
      if (log) print("mqtt can't resub $topic when force=false");

      return futureWaitData ? com.future : Future.value();
    }
    var id = getFreeID();
    _idTopic[id] = [topic];

    var msg = SubscribePacket.withTopic(id, topic);
    msg.head.retain = retain;
    msg.head.qos = qos;
    msg.head.dup = dup;
    _subList[topic] = msg;
    // handle future when suback
    if (!futureWaitData) {
      _subComplate[topic] = () {
        if (!com.isCompleted) {
          _timeout?.cancel();
          com.complete();
          _subComplate.remove(topic);
        }
      };
    }
    _send(msg);
    return com.future;
  }

  Future<void> unSubscribe(String topic) async {
    // print("mqtt::wsUnsub: $topic");
    // _subCache.remove(topic);
    // _subList.where((e) => e.topics.contains(topic))
    if (log) print("mqtt: unsub $topic");

    _subList.remove(topic);
    _dataArriveCallBack.remove(topic);
    _finishedSubCache.remove(topic);
    _send(UnSubscribePacket.withTopic([topic]));
    return;
  }

  /// internal function
  MqttClient._(
    String _host,
    int _port, {
    required this.credentials,
    required this.log,
    required this.allowReconnect,
    required this.reconnectWait,
    required this.customReconnectDelayCB,
  }) {
    _conn = XTransportTcpClient.from(_host, _port, credentials: credentials);
  }

  MqttClient(
    String _host,
    int _port, {
    XtransportCredentials credentials = const XtransportCredentials.insecure(),
    bool log = false,
    // reconnect attributes
    bool allowReconnect = false,
    Duration reconnectWait = const Duration(seconds: 2),
    Duration Function()? customReconnectDelayCB,
  }) : this._(
          _host,
          _port,
          credentials: credentials,
          allowReconnect: allowReconnect,
          reconnectWait: reconnectWait,
          customReconnectDelayCB: customReconnectDelayCB,
          log: log,
        );

  /// pause (ex: when app in backgroud mode)
  void pause() {
    _paused = true;
    _conn.close();
  }

  /// resume (ex:when app resume to fourcegroud mode)
  void resume() {
    _paused = false;
  }

  /// close the connect
  void close() {
    _conn.close();
  }

  void _onClose() {
    if (_stoped) return; //return if stoped
    if (_paused) {
      //check 1 second later if paused
      Future.delayed(Duration(seconds: 1)).then((_) => _onClose());
      return;
    }
    // if reconnectDuration not null, reconnect
    if (allowReconnect) {
      Future.delayed(customReconnectDelayCB?.call() ?? reconnectWait).then((_) {
        // _onBeforeReconnect?.call().th;
        if (_onBeforeReconnect != null) {
          return _onBeforeReconnect?.call().then((value) => _conn.connect());
        } else {
          _conn.connect();
        }
      });
    }
  }

  void _resetTimePeriodic() {
    var _seconds = _connectPacket.getKeepalive();
    _pinger?.cancel();
    if (_seconds > 0) {
      _pinger = Timer.periodic(Duration(seconds: _seconds), (timer) {
        _send(PingPacket());
      });
    }
  }

  void _send(ITransportPacket obj) {
    if (log) print("mqtt:ready send $obj");

    if (_stoped) return;
    if (_paused) return;
    if (_conn.status != ConnectStatus.connected) return;
    _resetTimePeriodic();
    _conn.send(obj);
    if (log) print("mqtt:sended $obj");
  }

  void reSub() {
    if (log) print("mqtt:reSub: ${_subList.keys.toList()}");

    _subList.forEach((key, value) {
      var id = getFreeID();
      _idTopic[id] = _subList[key]!.topics;
      _subList[key]!.withMessageID(id);
      _send(_subList[key]!);
    });
  }

  /// start connect
  void start() {
    if (_stoped) return;
    if (_started) return; //once function
    _started = true;
    //register events
    _conn.onConnect(() {
      _idTopic.clear();
      _finishedSubCache.clear();
      resetFreeID();
      _send(_connectPacket);
    });
    _conn.onClose(() {
      _onClose();
    });
    _conn.onMessage((msg) {
      if (_stoped) return;
      if (_paused) return;
      _resetTimePeriodic();
      var buf = MqttByteBuffer.fromList(msg.message);
      var head = MqttHeader.fromByteBuffer(buf);
      if (log) print("onHead: ${head.toString()}");
      switch (head.type) {
        case MqttMessageType.CONNACK:
          final obj = ConackPacket.fromByteBuffer(head, buf);
          _onMqttConack?.call(obj);
          if (_onMqttConack == null &&
              obj.returnCode != MqttConnectReturnCode.connectionAccepted) {
            _conn.close();
          }
          break;
        case MqttMessageType.SUBACK:
          var obj = SubackPacket.fromByteBuffer(head, buf);
          if (_idTopic.containsKey(obj.msgID)) {
            var _topics = _idTopic[obj.msgID]!;
            // print("this topic: $_topics");
            _idTopic.remove(obj.msgID);
            for (var _topic in _topics) {
              if (log) print('mqtt::sub confirmed for topic $_topic');
              // print('id map: $_idTopic');
              _finishedSubCache.add(_topic);
              // var com = _subComplate[_topic];
              // if (com?.isCompleted == false) {
              //   com?.complete();
              // }
              _subComplate[_topic]?.call();
            }
          }
          break;
        case MqttMessageType.PUBLISH:
          var obj = PublishPacket.fromByteBuffer(head, buf);
          _dataArriveCallBack[obj.topicName]?.call(obj);
          break;
        default:
      }
    });
    _conn.connect();
  }

  /// stop this client
  void stop() {
    _stoped = true;
    pause();
  }
}
