import 'dart:typed_data';

class MqttByteBuffer {
  /// The byte buffer
  // MqttByteBuffer(this.buffer);

  MqttByteBuffer() {
    // Uint8List().removeRange(start, end)
    buffer = <int>[];
  }

  /// From a list
  MqttByteBuffer.fromList(Iterable<int> data) {
    buffer = [];
    buffer.addAll(data);
  }

  /// The current position within the buffer.
  int _position = 0;

  /// The underlying byte buffer
  late List<int> buffer;

  /// Position
  int get position => _position;

  /// Length
  int get length => buffer.length;

  /// Available bytes
  int get availableBytes => length - _position;

  /// Resets the position to 0
  void reset() {
    _position = 0;
  }

  /// Skip bytes
  set skipBytes(int bytes) => _position += bytes;

  /// Add a list
  void addAll(List<int> data) {
    buffer.addAll(data);
  }

  /// Shrink the buffer
  void shrink() {
    buffer.removeRange(0, _position);
    _position = 0;
  }

  /// Message available
  bool isMessageAvailable() {
    if (availableBytes > 0) {
      return true;
    }

    return false;
  }

  /// Reads a byte from the buffer and advances the position
  /// within the buffer by one byte, or returns -1 if at the end of the buffer.
  int readByte() {
    final tmp = buffer[_position];
    if (_position <= (length - 1)) {
      _position++;
    } else {
      return -1;
    }
    return tmp;
  }

  /// Read a short int(16 bits)
  int readShort() {
    final high = readByte();
    final low = readByte();
    return (high << 8) + low;
  }

  /// Reads a sequence of bytes from the current
  /// buffer and advances the position within the buffer
  /// by the number of bytes read.
  Uint8List read(int count) {
    if ((length < count) || (_position + count) > length) {
      throw Exception('mqtt_client::ByteBuffer: The buffer did not have '
          'enough bytes for the read operation '
          'length $length, count $count, position $_position, buffer $buffer');
    }
    final tmp = <int>[];
    tmp.addAll(buffer.getRange(_position, _position + count));
    _position += count;
    return Uint8List.fromList(tmp);
  }

  /// Writes a byte to the current position in the buffer
  /// and advances the position within the buffer by one byte.
  void writeByte(int byte) {
    if (buffer.length == _position) {
      buffer.add(byte);
    } else {
      buffer[_position] = byte;
    }
    _position++;
  }

  /// Write a short(16 bit)
  void writeShort(int short) {
    writeByte(short >> 8);
    writeByte(short & 0xFF);
  }

  /// Writes a sequence of bytes to the current
  /// buffer and advances the position within the buffer by the number of
  /// bytes written.
  void write(Uint8List buffer) {
    this.buffer.addAll(buffer);
    _position = length;
  }

  /// Seek. Sets the position in the buffer. If overflow occurs
  /// the position is set to the end of the buffer.
  void seek(int seek) {
    if ((seek <= length) && (seek >= 0)) {
      _position = seek;
    } else {
      _position = length;
    }
  }

  /// Writes an MQTT string member
  void writeMqttUint8ListM(Uint8List stringToWrite) {
    writeMqttUint8List(this, stringToWrite);
  }

  /// Writes an MQTT string.
  /// stringStream - The stream containing the string to write.
  /// stringToWrite - The string to write.
  static void writeMqttUint8List(
      MqttByteBuffer stringStream, Uint8List stringToWrite) {
    final stringBytes = Uint8List(0);
    stringBytes.add(stringToWrite.length >> 8);
    stringBytes.add(stringToWrite.length & 0xFF);
    stringBytes.addAll(stringToWrite);
    stringStream.write(stringBytes);
  }

  // /// Writes an MQTT string member
  // void writeMqttStringM(String stringToWrite) {
  //   writeMqttString(this, stringToWrite);
  // }

  /// Writes an MQTT string.
  /// stringStream - The stream containing the string to write.
  /// stringToWrite - The string to write.
  // static void writeMqttString(
  //     MqttByteBuffer stringStream, String stringToWrite) {
  //   final enc = MqttEncoding();
  //   final stringBytes = enc.getBytes(stringToWrite);
  //   stringStream.write(stringBytes);
  // }

  /// Reads an MQTT string from the underlying stream member
  // String readMqttStringM() => MqttByteBuffer.readMqttString(this);

  // /// Reads an MQTT string from the underlying stream.
  // static String readMqttString(MqttByteBuffer buffer) {
  //   // Read and check the length
  //   final lengthBytes = buffer.read(2);
  //   final enc = MqttEncoding();
  //   final stringLength = enc.getCharCount(lengthBytes);
  //   final stringBuff = buffer.read(stringLength);
  //   return enc.getString(stringBuff);
  // }

  @override
  String toString() {
    if (buffer.isNotEmpty) {
      return 'null or empty';
    } else {
      return buffer.toList().toString();
    }
  }
}
